import os
import tqdm
import shutil
from . import utils
from urllib.request import urlopen
from bs4 import BeautifulSoup, SoupStrainer
from multiprocessing import Process, Queue
from typing import List


_NAV_ARTICLES_URL = ('http://news.naver.com/main/list.nhn?mode=LSD&mid=shm'
                     '&sid1={category}&date={date}&page={page}')


def _get_max_nav_pages(category: int,
                       date: str,
                       max_page: int = 100) -> int:
    nav_url = _NAV_ARTICLES_URL.format(category=category,
                                       date=date,
                                       page=max_page)

    # Read navigation page.
    with urlopen(nav_url) as res:
        # Use `SoupStrainer` to improve performance.
        strainer = SoupStrainer('div', {'class': 'paging'})
        document = BeautifulSoup(res.read(), 'lxml', parse_only=strainer)
        max_page_elem = document.find('strong')

    return int(max_page_elem.get_text())


def _get_article_urls_from_nav_page(category: int,
                                    date: str,
                                    page: int) -> List[str]:
    nav_url = _NAV_ARTICLES_URL.format(category=category,
                                       date=date,
                                       page=page)

    # Read navigation page and extract article links.
    with urlopen(nav_url) as res:
        # Use `SoupStrainer` to improve performance.
        strainer = SoupStrainer('div', {'class': 'list_body newsflash_body'})
        document = BeautifulSoup(res.read(), 'lxml', parse_only=strainer)

        article_items = document.find_all('li')
        article_urls = [item.find('a').get('href') for item in article_items]

    return article_urls


def _get_article_content(article_url: str) -> str:
    with urlopen(article_url) as res:
        # Use `SoupStrainer` to improve performance.
        strainer = SoupStrainer('div', attrs={'id': 'articleBodyContents'})
        document = BeautifulSoup(res.read(), 'lxml', parse_only=strainer)
        content = document.find('div')

        # Skip if there is no article content in the page.
        if content is None:
            return ''

        # Remove unnecessary tags.
        for child in content.find_all():
            child.clear()

        article_content = ' '.join(content.get_text().split())

    return article_content


def _crawl_articles_worker(output_file: str,
                           article_urls: List[str],
                           queue: Queue):
    with open(output_file, 'w', encoding='utf-8') as fp:
        for article_url in article_urls:
            fp.write(_get_article_content(article_url) + '\n')
            queue.put(None)


def start_crawling_articles(output_file: str,
                            temporary: str,
                            num_cores: int,
                            category_list: List[str],
                            start_date: str,
                            end_date: str,
                            max_page: int = 100):
    """Crawl news articles in parallel.

    Arguments:
        output_file (str): Output file path.
        temporary (str): Temporary directory path.
        num_cores (int): The number of processes.
        category_list (list): The list of categories to crawl from.
        start_date (str): Start date string.
        end_date (str): End date string.
        max_page (int): The maximum pages to crawl.
    """
    date_list = utils.drange(start_date, end_date)
    total_search = len(date_list) * len(category_list)
    tqdm_iter = tqdm.trange(total_search, desc='[*] collect article urls')

    article_urls = []
    for category in category_list:
        for date in date_list:
            # Get maximum number of pages in navigation.
            pages = _get_max_nav_pages(category, date, max_page)

            for page in range(1, pages + 1):
                article_urls += _get_article_urls_from_nav_page(
                    category, date, page)

            # Update progress bar.
            tqdm_iter.update()
    tqdm_iter.close()

    print(f'[*] successfully collecting article urls.'
          f' total articles: {len(article_urls)}')

    # Prepare multi-processing.
    workers = []
    queue = Queue()

    # Create temporary files and split articles into chunks.
    crawled_files = utils.random_filenames(temporary, num_cores)
    article_list_chunks = utils.split_list(article_urls, chunks=num_cores)

    for i in range(num_cores):
        w = Process(target=_crawl_articles_worker,
                    args=(crawled_files[i], article_list_chunks[i], queue))
        w.daemon = True
        w.start()

        workers.append(w)

    # Show crawling progress.
    for _ in tqdm.trange(len(article_urls), desc='[*] crawl articles'):
        queue.get(block=True)

    # Merge temporary files into `output_file`.
    print(f'[*] finish crawling articles. merge chunks into [{output_file}].')
    with open(output_file, 'wb') as dst:
        for name in crawled_files:
            with open(name, 'rb') as src:
                shutil.copyfileobj(src, dst)

    # Remove temporary files.
    for name in crawled_files:
        os.remove(name)



import os
import random
import string
from datetime import datetime, timedelta
from typing import List


_CANDIDATES = string.digits + string.ascii_lowercase


def drange(start: str, end: str) -> List[str]:
    """Return dates between ``start`` and ``end``.

    Arguments:
        start (str): Start date string.
        end (str): End date string.

    Note:
        Each date string must be in `yyyyMMdd` format.

    Returns:
        A list of date strings.
    """
    start_date = datetime.strptime(start, '%Y%m%d')
    end_date = datetime.strptime(end, '%Y%m%d')

    ndays = (end_date - start_date).days
    return [(start_date + timedelta(days=d)).strftime('%Y%m%d')
            for d in range(ndays + 1)]


def split_list(seq: List, chunks: int):
    """Split list into chunks.

    Arguments:
        seq (list): The list to split.
        chunks (int): The number of chunks.

    Returns:
        A list of chunks which are from input list.
    """
    chunk_list = []
    chunk_size = len(seq) // chunks

    for i in range(chunks - 1):
        chunk_list.append(seq[i * chunk_size:(i + 1) * chunk_size])
    chunk_list.append(seq[(chunks - 1) * chunk_size:])

    return chunk_list


def random_filename(parent: str) -> str:
    r"""Generate random filename consists of 16 characters.

    Arguments:
        parent (str): Parent directory for file.

    Returns:
        Randomly generated file name.
    """
    return os.path.join(parent, ''.join(random.choices(_CANDIDATES, k=16)))


def random_filenames(parent: str, n: int) -> List[str]:
    r"""Generate multiple random filenames.

    Arguments:
        parent (str): Parent directory for files.
        n (int): Total number of filenames.

    Returns:
        Randomly generated file names.
    """
    return [random_filename(parent) for _ in range(n)]



import os
import shutil
import argparse
from .crawling import start_crawling_articles


def _main():
    parser = argparse.ArgumentParser(
        prog='canrevan', description='crawl news article.')
    parser.add_argument('--output_file',
                        default='articles.txt',
                        help='output file path')
    parser.add_argument('--temporary',
                        default='tmp',
                        help='temporary directory path')
    parser.add_argument('--num_cores',
                        default=4,
                        type=int,
                        help='number of processes')
    parser.add_argument('--category',
                        required=True,
                        nargs='*',
                        type=int,
                        help='categories to crawl from')
    parser.add_argument('--start',
                        required=True,
                        help='start date string')
    parser.add_argument('--end',
                        required=True,
                        help='end date string')
    parser.add_argument('--max_page',
                        default=100,
                        type=int,
                        help='maximum number of pages to search')
    args = parser.parse_args()

    # Create temporary directory.
    if not os.path.exists(args.temporary):
        os.makedirs(args.temporary)

    # Crawl news articles.
    start_crawling_articles(args.output_file,
                            args.temporary,
                            args.num_cores,
                            args.category,
                            args.start,
                            args.end,
                            args.max_page)

    # Remove temporary directory.
    shutil.rmtree(args.temporary)
